option show_stats;
option solver gurobi_ampl;

## Model parameters
param p; # Number of processors
param b; # Child per node
param L = log(p) / log(b);
set F;
set FF = union{i in F} { i * 1 };

## Optimization
param M; # Crown maximal makespan
param MM = M;
param zeta; # Weight for static energy consumption
param kappa; # Constant part of static power
param eta; # Idle power

set P = 1..p; # Set of processors
set B = 1..b; # Set of child at each node except leaves
param q = 2 * p - 1;

## Hierarchical structure of processor groups
set Pi{i in 1..q} within P = if floor(log(i) / log(b)) + 1 <= L then union{j in B} Pi[b * i + j - 1] else {i - b^L + 1};
set G{i in P} = union{j in 1..q} if i in Pi[j] then {j} else {};
param pi{i in 1..q} = card(Pi[i]);
# Matrix equivalent to pi (MM + pi = Mpi)
param Mpi{i in 1..q, j in P} = if pi[i] = j then 1 else 0;

## Tasks
param n; # Number of tasks
set T = 1..n; # Task set
param Wi{i in T}; # Task width
param Tau{i in T}; # Task work
param e{i in T, j in P}; # Task efficiency distribution
param c{i in T, j in T}; # Task channels

## FFrequency
param alpha; # FFrequency cost parameter
param fa{i in FF} = (zeta * i^(alpha) + (1 - zeta) * (i + kappa * min{j in FF}j)) / zeta; # Cost of a frequency
 
## Integrated mapping
var x{i in 1..q, j in T, k in FF} binary; # 3D Matrix for frequency/group mapping
subject to no_mapping_to_bigger_groups{j in T}: sum{i in 1..(p / (b^floor(log(min(Wi[j], p)) / log(2))) - 1)} sum{k in FF} x[i, j, k] = 0; # A task cannot be mapped to a group larger than its maximum width
subject to one_setting_per_task{j in T}: sum{i in p / (b^floor(log(min(Wi[j], p)) / log(2)))..q} sum{k in FF} x[i, j, k] = 1; # A task must be mapped to one frequency and one group whose size is at most as large than its maximum width
## Debug only: check if there is any energy loss by fordidding a group to run more than a task. For sequential tasks sets, means that task would be spread out hrough the cores.
#subject to one_task_per_group{i in 1..q}: sum{j in T} sum{k in FF} x[i,j,k] <= 1;

## More variables to ease reading of x
var grp{j in T} = sum{i in 1..q} sum{k in FF} x[i, j, k] * i;
var freq{j in T} = sum{i in 1..q} sum{k in FF} x[i, j, k] * k;
var _int_frequency{i in T} = freq[i]; # Alias for output with non-confusing name
var frequency{i in T} = round(_int_frequency[i]); # make sure output is integer digits
var _int_group{i in T} = grp[i]; # Alias for output with non-confusing name
var group{i in T} = round(_int_group[i]); # make sure output is integer digits

## Time
var time{ii in P} = sum{i in G[ii]} sum{l in P} Mpi[i, l] * sum{k in FF} sum{j in T} (x[i, j, k] * Tau[j] / (e[j, l] * l * k));
subject to throughput{i in P}: time[i] <= MM;
var time_ok{ii in P} = if time[ii] <= MM then 1 else 0;

var int_time{i in P} = time[i]; # Alias for output with non-confusing name

## Set to 1 if processor ii is used at all
var u{ii in P} binary;
subject to one_if_used{ii in P}: u[ii] <= sum{i in G[ii]} sum{j in T} sum{k in FF} x[i,j,k];
subject to zero_if_unused{ii in P, i in G[ii], j in T, k in FF}: u[ii] >= x[i,j,k];

## Energy: static, dynamic and weighted global
var time_busy = sum{i in 1..q} sum{l in P} Mpi[i, l] * sum{k in FF} sum{j in T} x[i, j, k] * Tau[j] / (e[j, l] * k);
var int_busy = sum{i in 1..q} sum{l in P} Mpi[i, l] * sum{k in FF} fa[k] * sum{j in T} x[i, j, k] * Tau[j] / (e[j, l] * k);
var time_idle_active = ((MM * (card(P) - sum{i in P} (1 - u[i]))) - time_busy);
var int_idle = time_idle_active * eta * min{j in FF}fa[j];
var int_energy = (zeta * int_busy + (1 - zeta) * int_idle);

## Give allocation output
var wi{i in T} = p / (b ^ floor((log(round(grp[i])) / log(b))));

## Distribute tasks so we reduce the makespan of a schedule
## Actually distribute the work each core is given, where frequency has no incidence; this prevent teh solver from trying to optimize makespan by running tasks faster
## Also prevent the solver from trying to reduce the makespan by parallelizing tasks because it can make it to sacrify energy for better makespan. Assumes parallelized tasks run several times sequentially on all processors and try to balance the resulting load
## Make sure the efficency loss due parallelization choice doesn't create different runtime ratio between tasks when they are parallelized and not parallelized. Penalize a sequential task's load with the efficiency of the chosen parallelization makes sure this ratio is preserved even if all tasks have a different efficiency function, but assuming this function is decreasing from 1 to +infinite.
## In other words, gets rid of values (parallelization time + frequency scaling) that can tempt the solver to modify its solution to achieve a better makespan, but make sure the parallelized work order relationship between tasks is preserved with the work loads we are trying to balance.
var core_work_eq{ii in P} = sum{i in G[ii]} sum{l in P} Mpi[i, l] * sum{k in FF} sum{j in T} (x[i, j, k] * Tau[j] / e[j, l]); ## core work equivalent (equivalent = different value than work, but having the same ratio as between each tasks and cores' work after parallelization)
var max_core_work_eq;
subject to balance_core_work_eq{i in P}: core_work_eq[i] <= max_core_work_eq;

## Wrap energy optimization and task balance together

var energy_balance = int_energy; ## Only take energy efficiency into account
var used_cores = card(P) - sum{i in P} (1 - u[i]);
var time_task{j in T} = sum{i in 1..q} sum{l in P} Mpi[i, l] * sum{k in FF} x[i, j, k] * Tau[j] / (l * e[j, l] * freq[j]);

minimize target: energy_balance;
