# Copyright 2015 Nicolas Melot
#
# This file is part of Crown.
#
# Crown is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Crown is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Crown. If not, see <http://www.gnu.org/licenses/>.
#


option show_stats;
option solver gurobi_ampl;

## Model parameters
param p; # Number of processors (log2)
param b; # Child per node
param L = log(p) / log(b); # Number of processors
param epsilon;

### Useless parameters, but still given as input
param zeta;
param M;
set F;
param alpha;

set P = 1..p; # Set of processors
set B = 1..b; # Set of child at each node except leaves

## Convenient variables
param pob{i in P} = if log(i) / log(b) = floor(log(i) / log(b)) then 1 else 0; # 1 if i is a power Of B
param q = 2 * p -1; # max number of cores in a group
param group{i in P} = b ^ ceil(log(i) / log(b)); # To what group would be mapped a non-power-of-b number of cores

## Hierarchical structure of processor groups
set Pi{i in 1..q} within P = if floor(log(i) / log(b)) + 1 <= L then union{j in B} Pi[b * i + j - 1] else {i - b^L + 1};
set G{i in P} = union{j in 1..q} if i in Pi[j] then {j} else {};
param pi{i in 1..q} = card(Pi[i]);
# Matrix equivalent to pi (M + pi = Mpi)
param Mpi{i in 1..q, j in P} = if pi[i] = j then 1 else 0;

## Tasks
param n; # Number of tasks
set T = 1..n; # Task set
param Wi{i in T}; # Task width
param Tau{i in T}; # Task work
param e{i in T, j in P}; # Task efficiency distribution
param e_max{i in T, j in P} = if Wi[i] >= j then e[i, j] else min(e[i, j], (e[i, j] * Wi[i]) / j); # Maximal task efficiency after a task's given efficiency and width

param test{i in P, j in T} = i + j;

## Resource allocation
var _wi{i in T};
subject to min_cores {i in T}: _wi[i] >= 1; # At least one core per task
subject to max_cores {i in T}: _wi[i] <= max(1, min(Wi[i], p)); # Max p core to task
var wi{i in T} = round(_wi[i]); # make sure output is integer digits

## Matrix equivalent to wi (M + wi = Mwi)
var Mwi{i in T, j in P} binary;
subject to exactly_one_map{i in T}: sum{j in P} Mwi[i, j] = 1; # A task get exactly one number of procs
subject to map_Mwi_to_wi{i in T}: sum{j in P} Mwi[i, j] * j = _wi[i]; # Number of cores given to task matches wi.
subject to power_of_b{i in T}: sum{j in P} Mwi[i, j] * pob[j] = 1; # Each task must be allocated a power of b cores.

## Parallel work for tasks mapped to groups
var pTau{i in T} = sum{j in P} Mwi[i, j] * Tau[i] / j * e_max[i ,j];
var Rz{i in P} = sum{j in T} Mwi[j, i] * pTau[j] / (p / i);
var R = sum{i in P} Rz[i];

minimize allocation: R;
