# Copyright 2015 Nicolas Melot
#
# This file is part of Crown.
#
# Crown is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Crown is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Crown. If not, see <http://www.gnu.org/licenses/>.
#


option show_stats;
option solver gurobi_ampl;

## Model parameters
param p; # Number of processors
param b; # Child per node
param L = log(p) / log(b);
param M;

## Cost parameters
param zeta;
param epsilon;

set P = 1..p; # Set of processors
set B = 1..b; # Set of child at each node except leaves

## Convenient variables
param q = 2 * p -1;

## Hierarchical structure of processor groups
set Pi{i in 1..q} within P = if floor(log(i) / log(b)) + 1 <= L then union{j in B} Pi[b * i + j - 1] else {i - b^L + 1};
set G{i in P} = union{j in 1..q} if i in Pi[j] then {j} else {};
param MG{i in P, j in 1..q} = if i in Pi[j] then 1 else 0;

param pi{i in 1..q} = card(Pi[i]);
# Matrix equivalent to pi (M + pi = Mpi)
param Mpi{i in 1..q, j in P} = if pi[i] = j then 1 else 0;

## Tasks
param n; # Number of tasks
set T = 1..n; # Task set
param Wi{i in T}; # Task width
param Tau{i in T}; # Task work
param e{i in T, j in P}; # Task efficiency distribution
#param e_max{i in T, j in P} = if Wi[i] >= j then e[i, j] else min(e[i, j], min(e[i, Wi[i]], Wi[i] / j)); # Maximal task efficiency after a task's given efficiency and width

## Resource allocation
#param wi{i in T} = max(1, min(Wi[i], p)); # Always give maximal resources a task can handle
param wi{i in T}; # Get resource allocation from mapping.dat
# Matrix equivalent to wi (M + wi = Mwi)
param Mwi{i in T, j in P} binary = if wi[i] = j then 1 else 0; # Important warning: j in P stands here for the number of cores allocated to task i, it is NOT the processor it's been mapped to.

## Task mapping to group
#param group{i in T} within 1..q = b ^ (L - ceil(log(wi[i]) / log(b))); # We give task i the smallest group able to provide all resources allocated for i
param group{i in T} within 1..q; # Load task mapping from frequency.dat
param _step_group{i in T} = group[i];
param step_group{i in T} = round(_step_group[i]); # make sure output is integer digits
param y{i in T, j in 1..q} = if group[i] = j then 1 else 0; # Matrix equivalent for group

## Frequency
set F;
param alpha;
param fa{i in F} = i^(alpha) + zeta * i; # Cost of a frequency

## Task-Frequency matrix
var f{i in T, j in F} binary;
subject to exactly_one_frequency{i in T}: sum{j in F} f[i, j] = 1; 
var freq{i in T}; # Frequency mapped to task i
subject to map_f_to_frequency{i in T}: freq[i] = sum{j in F} f[i, j] * j;
var _step_frequency{i in T} = freq[i]; # Alias for output with non-confusing name
var frequency{i in T} = round(_step_frequency[i]); # make sure output is integer digits

## Time
#var time{i in P} = sum{j in T} sum{k in F} (f[j, k] * y[j, i] * Tau[j] / k * wi[j] * sum{l in P} Mwi[j, l] * e_max[j, l]);
#var time{i in P} = sum{j in T} sum{k in F} sum{l in 1..q} MG[i, l] * (f[j, k] * y[j, l] * Tau[j] / (k * wi[j] * sum{m in P} Mwi[j, m] * e_max[j, m]));
var time{i in P} = sum{j in T} sum{k in F} sum{l in 1..q} MG[i, l] * (f[j, k] * y[j, l] * Tau[j] / (k * wi[j] * sum{m in P} Mwi[j, m] * e[j, m]));
subject to throughput {i in P}: time[i] <= M;
var step_time{i in P} = time[i]; # Alias for output with non-confusing name

## Energy
var step_busy = sum{i in T} sum{k in P} Mwi[i, k] * sum{j in F} f[i, j] * fa[j] * Tau[i] / e[i, k];
var step_idle = sum{j in P} (M - time[j]) * min{i in F} fa[i];
var step_energy = step_busy + epsilon * step_idle;

minimize voltage: step_energy;
